<?php

/*

 rn_xmlrpc.php : An XML-RPC publishing API Implimentation for rNote.

 Copyright (C) 2005 Tousei (冬星)
 http://ichounoki.info/rnote/

 Version 1.1.2.3

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

require_once("XML/RPC/Server.php");
require_once("rnote_config.php");
require_once("rnote_lib.php");

define(FOPEN_USE_FTP,   false);

//==================================================================================
// Defines and Global vars.
// 定数定義とグローバル
//==================================================================================

// UTF-8 設定。
$GLOBALS['XML_RPC_defencoding'] = "UTF-8";
mb_internal_encoding('UTF-8');

//XML_RPC_Response のユーザー定義エラーコード
define(XR_ERR_ARG,   1); //パラメータが不正。
define(XR_ERR_USER,  2); //ユーザーアカウントand/orパスワードが不正。
define(XR_ERR_POSTID,3); //posidが不正。
define(XR_ERR_FUNC,  4); //サーバー側処理で起きたエラー。
//
define(YEAR_FMT,'Y');    //年の表現のフォーマット指定子。

define(FLG_NR,1); // 自動改行の方式。0: <p></p>に置き換える。1: <br /> に置き換える。
define(FLG_YR,1); // 西暦表記。'0': 2桁で表示、1: 4桁で表示。
define(FNAME_FILECACHE,DIR_CACHE.'file.cache');
define(FNAME_DIRCACHE,DIR_CACHE.'dir.cache');

//ユーザー情報中のオプションに設定される要素。
define(XR_OPT_FORCE_CONVERT_BREAKS,'force_convert_breaks'); //強制的に改行置き換え処理する。
define(XR_OPT_FORCE_ALLOW_COMMENTS,'force_allow_comments'); //強制的にコメントを許可とする。
define(XR_OPT_FORCE_PUBLISH,'force_publish'); //強制的に newPost() 直後に再構築する。

//==================================================================================
// Functions for XML-RPC portability.
//グローバル変数の宣言に関してはこうしたラッパーが使えないので移植時は以下に注意。
//修正すべき変数名：$XML_RPC_erruser,$xmlrpcerruser
//==================================================================================

//xrErrResp --- fault 処理。
//[戻り値]
function xrErrResp($dummy, $code, $add)
{
    $msg = '';
    switch ($code) {
    case XR_ERR_ARG:
        $msg = 'Invalid argument';
        break;
    case XR_ERR_USER:
        $msg = 'Invalid user';
        break;
    case XR_ERR_POSTID:
        $msg = 'Invalid postid';
        break;
    case XR_ERR_FUNC:
        $msg = 'Function eror';
        break;
    }
    //PEAR
    return new XML_RPC_Response($dummy, $xmlrpcerruser+$code, $msg.": ".$add);
}

//xrResp --- 正常応答処理。
//[戻り値]
function xrResp($msg)
{
    //PEAR
    return new XML_RPC_Response($msg);
}

//xrChmod --- 所有権を FTP ユーザーにした chmod 関数。
//[戻り値]
//  成功: int fp 。失敗: FALSE。
function xrChmod($path, $pmt)
{
    if (!FOPEN_USE_FTP) {
        return chmod($path, $pmt);
    }
    // FTP を用いて所有者を FTP ユーザーにする 。
    $is_ftp_success = FALSE;
    $fs = ftp_connect(XR_FTP_HOST);
    if (ftp_login($fs, XR_FTP_USER, XR_FTP_PW)) {
        ftp_pwd($fs);
        if (ftp_chdir($fs, XR_FTP_DIR)) {
            $m = sprintf("%03o", $pmt);
            $chmod_cmd = "CHMOD ".$m." ".$path;
            if (ftp_site($fs, $chmod_cmd)) {
                $is_ftp_success = TRUE;
                WriteLog("FTP command CHMOD OK: path=[".$path."]pmt=[".$m."]\n");
            }
            //if (ftp_chmod($path, $pmt)) {
            //    $is_ftp_success = TRUE;
            //}
        }
    }
    ftp_quit($fs);
    return is_ftp_success;
}
//xrMkdir --- 所有権を FTP ユーザーにした mkdir 関数。
//[戻り値]
//  成功: int fp 。失敗: FALSE。
function xrMkdir($path)
{
    if (!FOPEN_USE_FTP) {
        return mkdir($path);
    }
    //既に存在していればそのまま。
    clearstatcache();
    if (is_dir($path)) {
        WriteLog("path[".$path."] is already exist.\n");
        return TRUE;
    }
    // FTP を用いて所有者を FTP ユーザーにする 。
    $is_ftp_success = FALSE;
    $fs = ftp_connect(XR_FTP_HOST);
    if (ftp_login($fs, XR_FTP_USER, XR_FTP_PW)) {
        ftp_pwd($fs);
        if (ftp_chdir($fs, XR_FTP_DIR)) {
            if (ftp_mkdir($fs, $path)) {
                $is_ftp_success = TRUE;
                WriteLog("FTP command ftp_mkdir OK: path=[".$path."]\n");
            }
        }
    }
    ftp_quit($fs);
    return is_ftp_success;
}
//xrUnlink --- 所有権を FTP ユーザーにした unlink 関数。
//[戻り値]
//  成功: int fp 。失敗: FALSE。
function xrUnlink($path)
{
    if (!FOPEN_USE_FTP) {
        return unlink($path);
    }
    // FTP を用いて所有者を FTP ユーザーにする 。
    $is_ftp_success = FALSE;
    $fs = ftp_connect(XR_FTP_HOST);
    if (ftp_login($fs, XR_FTP_USER, XR_FTP_PW)) {
        ftp_pwd($fs);
        if (ftp_chdir($fs, XR_FTP_DIR)) {
            //$path は rn_xmlrpc.php があるディレクトリからの相対ディレクトリ。
            if (ftp_delete($fs, $path)) {
                $is_ftp_success = TRUE;
                WriteLog("FTP command ftp_delete OK: path=[".$path."]\n");
            }
        }
    }
    ftp_quit($fs);
    return is_ftp_success;
}
function xrFile_exists($path)
{
    WriteLog("---\nIn xrFile_exists:path=[".$path."]\n");
    //if (!FOPEN_USE_FTP) {
        clearstatcache();
        return file_exists($path);
    //}
    // FTP を用いて所有者を FTP ユーザーにする 。
    $is_ftp_success = FALSE;
    $fs = ftp_connect(XR_FTP_HOST);
    if (ftp_login($fs, XR_FTP_USER, XR_FTP_PW)) {
        ftp_pwd($fs);
        //全てのサーバーが対応しているわけではないそうだ。
        if (ftp_chdir($fs, XR_FTP_DIR)) {
            if (ftp_size($fs, $path) >= 0) {
                $is_ftp_success = TRUE;
                WriteLog("FTP command chmod OK: path=[".$path."]\n");
            }
        }
    }
    ftp_quit($fs);
    return is_ftp_success;
}

//xrFopen --- 所有権を FTP ユーザーにしたファイルオープン関数。
//  $pmt: パーミッションを設定する場合の値。 0 のときは設定しない。
//[戻り値]
//  成功: int fp 。失敗: FALSE。
function xrFopen($path, $mode, $pmt)
{
    WriteLog("---\nIn ftp_fput:\n");
    // Read の時はファイルが既に存在することが前提なのでそのまま fopen() する。
    if ((!FOPEN_USE_FTP) || ($mode == 'r') || ($mode == 'rb')) {
        $fp = @fopen($path, $mode);
        if ($fp && $pmt) {
            xrChmod($path, $pmt);
        }
        return $fp;
    }
    //既にファイルが存在する場合は FTP を用いることができない。
    //（ダミーファイルを PUT しようにも、既にファイルが存在するから。）
    clearstatcache();
    if (file_exists($path)) {
        $fp = @fopen($path, $mode);
        if ($fp && $pmt) {
            xrChmod($path, $pmt);
        }
        return $fp;
    }
    // FTP を用いて所有者を FTP ユーザーにする fopen() 。
    //
    $lockfilename = 'lock/lock_ftp';
    $lockfp = @fopen($lockfilename,'w'); //ロックファイルでロックする。
    if (!$lockfp) return FALSE;
    flock($lockfp, LOCK_EX);
    //-------------------------------------------
    $is_ftp_success = FALSE;
    $fs = ftp_connect(XR_FTP_HOST);
    if (ftp_login($fs, XR_FTP_USER, XR_FTP_PW)) {
        WriteLog("ftp_login OK\n");
        ftp_pwd($fs);
        ftp_chdir($fs, XR_FTP_DIR);
        $dummypath = XR_FTP_DUMMY_DIR . XR_FTP_DUMMY_FNAME;
        WriteLog("ftp_fput:[".$path."]<-[".$dummypath."]\n");
        //ダミーファイルは apache でも開くことができるような権利にしておくこと。
        if ($fp = @fopen($dummypath, 'rb')) {
            WriteLog("fopen [".$dummypath."] OK\n");
            //ダミーファイルから作りたいパスにコピーする。
            if (ftp_fput($fs, $path, $fp, FTP_BINARY)) {
                WriteLog("ftp_fput OK\n");
                if ($pmt) {
                    if (xrChmod($path, $pmt)) {
                        $is_ftp_success = TRUE;
                        WriteLog("FTP ALL OK\n");
                    }
                }
            }
            fclose($fp);
        }
    }
    ftp_quit($fs);
    //-------------------------------------------
    flock($lockfp, LOCK_UN); fclose($lockfp); //ロックファイルを閉じる。
    if (!is_ftp_success) {
        WriteLog("is_ftp_success false\n");
        return FALSE;
    }
    //FTPでファイルを用意したら、そのファイルをすぐに開く。
    $fp = fopen($path, $mode);
    WriteLog("Out ftp_fput:fp=[".$fp."], path="."./".$path."]mode=[".$mode."]file_exist=[".xrFile_exists("./".$path)."]\n");
    return $fp;
}
//xrFile_get_contents --- 所有権を FTP ユーザーにしたファイルオープン関数。
//[戻り値]
//  成功: string ファイルの内容 。失敗: FALSE。
function xrFile_get_contents($path)
{
    //if (!FOPEN_USE_FTP) {
        return @file_get_contents($path);
    //}
    if ($fp = xrFopen($path, 'rb', 0)) {
        $contents = fread($fp, filesize($path));
        fclose($fp);
        return $contents;
    }
    return FALSE;
}

//userCheck --- 許可されたユーザーかチェックする。
//[戻り値]
//  成功: user情報構造体。失敗: FALSE。
function userCheck($nm, $pw)
{
    global $XR_USER;

    for ($i = 0; $i < count($XR_USER); $i++) {
        $str = $XR_USER[$i];
        if (!strcmp($str['username'],$nm) && !strcmp($str['password'],$pw)) {
            return $str;
        }
    }
    return FALSE;
}

function replaceRightBlacket($str)
{
    //$str = eregi_replace("&lt;!--.*--&gt;", "", $str);
    //$str = str_replace('&gt;','>',$str);

    return $str;
}

//createCategoryFile --- .category ファイルを作る。
//[戻り値]
//  成功: TRUE。失敗: FALSE。
function createCategoryFile($postid, $contents)
{
    WriteLog("---\nIn createCategoryFile:\n");
    $path = 'entries/draft/'.$postid.'.category';
    //
    $lockfilename = 'lock/lock_categoryfile';
    $lockfp = @fopen($lockfilename,'w'); //ロックファイルでロックする。
    if (!$lockfp) return FALSE;
    flock($lockfp, LOCK_EX);
    //-------------------------------------------
    $fp = xrFopen($path,"wb", 0666);
    if ($fp) {
        set_file_buffer($fp, 0);
        flock($fp, LOCK_EX);
        rewind($fp);
        fwrite($fp, $contents, strlen($contents));
        flock($fp, LOCK_UN);
        fclose($fp);
    }
    //-------------------------------------------
    flock($lockfp, LOCK_UN); fclose($lockfp); //ロックファイルを閉じる。
    WriteLog("Out createCategoryFile:\n");
    if (!$fp) return FALSE;
    return TRUE;
}

//readCategoryFile --- .category ファイルからカテゴリーキーを得る。
//[戻り値]
//  成功: 読み込んだ中身。失敗: FALSE。
function readCategoryFile($postid)
{
    $path = 'entries/draft/'.$postid.'.category';
    //
    $lockfilename = 'lock/lock_categoryfile';
    $lockfp = @fopen($lockfilename,'w'); //ロックファイルでロックする。
    if (!$lockfp) return FALSE;
    flock($lockfp, LOCK_EX);
    //-------------------------------------------
    $contents = xrFile_get_contents($path);
    //-------------------------------------------
    flock($lockfp, LOCK_UN); fclose($lockfp); //ロックファイルを閉じる。
    return $contents;
}

//removeTBurlFile --- .trackback ファイルを削除する。
//[戻り値]
//  成功: TRUE。失敗: FALSE。
function removeTBurlFile($postid)
{
WriteLog("---\nIn removeTBurlFile:\n");
    $path = 'entries/draft/'.$postid.'.trackback';
    //
    $lockfilename = 'lock/lock_tburlfile';
    $lockfp = @fopen($lockfilename,'w'); //ロックファイルでロックする。
    if (!$lockfp) return FALSE;
    flock($lockfp, LOCK_EX);
    //-------------------------------------------
    xrUnlink($path);
    //-------------------------------------------
    flock($lockfp, LOCK_UN); fclose($lockfp); //ロックファイルを閉じる。
WriteLog("Out createTBurlFile:\n");
    return TRUE;
}

//createTBurlFile --- .trackback ファイルを作る。
//[戻り値]
//  成功: TRUE。失敗: FALSE。
function createTBurlFile($postid, $title, $excerpt, $tb_urls)
{
    if (!isset($tb_urls)) return TRUE;
WriteLog("---\nIn createTBurlFile:\n");
    $path = 'entries/draft/'.$postid.'.trackback';
    //
    $lockfilename = 'lock/lock_tburlfile';
    $lockfp = @fopen($lockfilename,'w'); //ロックファイルでロックする。
    if (!$lockfp) return FALSE;
    flock($lockfp, LOCK_EX);
    //-------------------------------------------
    $fp = xrFopen($path,"wb", 0666);
    if ($fp) {
        set_file_buffer($fp, 0);
        flock($fp, LOCK_EX);
        rewind($fp);
        fwrite($fp, $title); fwrite($fp, "\r\n");
        fwrite($fp, $excerpt); fwrite($fp, "\r\n");
        foreach ($tb_urls as $url) {
            fwrite($fp, $url); fwrite($fp, "\r\n");
        }
        flock($fp, LOCK_UN);
        fclose($fp);
    }
    //-------------------------------------------
    flock($lockfp, LOCK_UN); fclose($lockfp); //ロックファイルを閉じる。
WriteLog("Out createTBurlFile:\n");
    if (!$fp) return FALSE;
    return TRUE;
}

//readTBurlFile --- .trackback ファイルからカテゴリーキーを得る。
//[戻り値]
//  成功: 読み込んだ中身。失敗: FALSE。
function readTBurlFile($postid, &$title, &$excerpt, &$tb_urls)
{
    $path = 'entries/draft/'.$postid.'.trackback';
    //
    $lockfilename = 'lock/lock_tburlfile';
    $lockfp = @fopen($lockfilename,'w'); //ロックファイルでロックする。
    if (!$lockfp) return FALSE;
    flock($lockfp, LOCK_EX);
    //-------------------------------------------
    $fp = xrFopen($path, "r", 0);
    if ($fp) {
        set_file_buffer($fp, 0);
        flock($fp, LOCK_EX);
        rewind($fp);
        if (!feof($fp)) {
            $title = fgets($fp, 4096);
            $title = str_replace("\r\n", '', $title);
        }
        if (!feof($fp)) {
            $excerpt = fgets($fp, 4096);
            $excerpt = str_replace("\r\n", '', $excerpt);
        }
        while (!feof($fp)) {
            $tb_url = fgets($fp, 4096);
            $tb_url = str_replace("\r\n", '', $tb_url);
            array_push($tb_urls, $tb_url);
        }
        flock($fp, LOCK_UN);
        fclose($fp);
    }
    //-------------------------------------------
    flock($lockfp, LOCK_UN); fclose($lockfp); //ロックファイルを閉じる。
    if (!count($tb_urls)) return FALSE;
    return TRUE;
}

function getall_XMLTag($tag,$str,&$array,$mode=PREG_SET_ORDER)
{
	return preg_match_all("/<$tag".'[^>]'."*?>[[:space:]]*?(.*?)[[:space:]]*?<\/$tag>/si",$str,$array,$mode);
}

//sendTrackBackPing --- Trackback ping URL リストが設定されていれば送信する。
function sendTrackBackPing($postid, $title, $excerpt, $tb_urls)
{
    $akey = getCategoryKeyFromPostid($postid);
    if (!$akey) return FALSE;
    $url = SITE_URL.FNAME_RNOTE.'?u='.$akey.'/'.$postid.'.htm';
    foreach ($tb_urls as $tb_url) {
        $msg = 'postid['.$postid.']url['.$url.'[Send Trackback ping to:'.$tb_url.']: ';
        //相手側に問い合わせて、既にトラックバック済みであれば送信しない。
        $ret = send_http($tb_url.'?__mode=rss', 'GET', '', '');
        if ($ret) {
            if (get_XMLTag('error',$ret)==='0') {
                // 1. item タグに括られていない情報として Trackback されている記事自体の情報が１件先頭にある。
                // 2. item タグに括られた内容が各 Trackback の情報なので、 1. の情報は捨てて item タグに
                //    括られた内容だけを求めるようにする。
                $c_i = getall_XMLTag('item',$ret,$items,PREG_SET_ORDER);
                for ($i=0;$i<$c_i;$i++) {
                    $c_l = getall_XMLTag('link',$items[$i][1],$links,PREG_SET_ORDER);
                    for ($j=0;$j<$c_l;$j++) {
                        if ($links[$j][1] == $url) {
                            //ALREADY SENT.
                            return TRUE;
                        }
                    }
                }
            }
        }

        $ret = send_http($tb_url, 'POST', '', 
                        array('url' => $url, 'blog_name' => SITE_TITLE, 'title' => $title, 
                              'excerpt' => $excerpt, 'charset' => 'utf-8'));
        if (get_XMLTag('error',$ret)!=='0') {
            //WriteLog($msg."sendTrackBackPing: NG\n");
            //WriteLog($ret);
            return FALSE;
        }
        //OK.
    }
    return TRUE;
}

//movePost --- 投稿記事を移動する。
//[戻り値]
//  成功 : TRUE。失敗 : FALSE。
function movePost($postid, $dst_cate, $src_cate)
{
    //移動元と移動先のパスをつくる。
    $src = DIR_DATA.$src_cate.'/'.$postid.'.xml';
    $dst = DIR_DATA.$dst_cate.'/'.$postid.'.xml';

    //移動元と移動先が同じカテゴリーなら何もしないで成功とする。
    //特に draft->draft の場合、draft に公開するということはしてはならないので .category も削除してはいけない。
    if ($dst_cate == $src_cate) return TRUE;

    $ret = FALSE;
    $lockfilename = 'lock/lock_post';
    $lockfp = @fopen($lockfilename,'w'); //ロックファイルでロックする。
    if (!$lockfp) return FALSE;
    flock($lockfp, LOCK_EX);
    //-------------------------------------------
    $contents = xrFile_get_contents($src);
    if ($contents) {
        //移動後の記事を Write モードで開く。
        $fp = xrFopen($dst, "wb", 0666);
        if ($fp) {
            set_file_buffer($fp, 0);
            flock($fp, LOCK_EX);
            rewind($fp);
            fwrite($fp, $contents, strlen($contents));
            flock($fp, LOCK_UN);
            fclose($fp);
        } else {
        }
        //書き込み終わったら、元のファイル（と、あれば .category ）を削除する。
        if (xrUnlink($src)) {
            //.category ファイルの処理。
            $path = 'entries/draft/'.$postid.'.category';
            if (is_file($path)) xrUnlink($path);
            $ret = TRUE;
        } else {
            //元のファイルが削除できないなら、移動後のファイルを削除してエラー終了する。
            xrUnlink($dst);
        }
    }
    //-------------------------------------------
    flock($lockfp, LOCK_UN); fclose($lockfp); //ロックファイルを閉じる。
    return $ret;
}

//nl2p --- 文字列の改行コードを <p></p> の組に置き換える。
//[引数]
//  $str : 置き換え対象の文字列。
//[戻り値]
//  置き換えられた文字列。
function nl2p($str)
{
    $str = str_replace("\r\n", "\r", $str); //for Win
    $str = str_replace("\r",   "\n", $str); //for Mac
    $str = str_replace("\n",   "</p>\n<p>", $str);
    $str = "<p>".$str."</p>\n";
    $end = "<p></p>\n";
    //一番最後が $end の終わり方なら末尾の $end 分は余分なので削除する。
    if (!strcmp(substr($str, -strlen($end)),$end)) {
        $str = substr($str, 0, strlen($str)-strlen($end));
    }
    return $str;
}

//createPost --- 投稿処理（新規投稿または更新投稿。）
//[引数]
//  $c_key : 書き込むべきカテゴリーのキー
//  $convert_breaks : 自動改行(1:する,0:しない)
//[戻り値]
//  成功: TRUE, 失敗: FALSE。
function createPost($postid, $c_key, $convert_breaks = 'none', $title, $description, 
                    $mt_text_more = '', $mt_allow_comments = 1, $mt_allow_pings = 1, $mt_excerpt = '', $mt_keywords = '')
{
    global $category_name;

    WriteLog("---\nIn createPost:\n");
    if ($postid) {
        //post id から更新処理の元ファイルのカテゴリーキーを作る。
        $oc_key = getCategoryKeyFromPostid($postid);
        if ((!$oc_key) || (!isset($postid))) return FALSE;
    }
    if ($title == "") $title = '(無題)';
    $description = trim($description); // 前後の余白を除去
    // 改行コードをタグに置き換えるかどうか。
    if (('convert_breaks' == $convert_breaks) || ('1' == $convert_breaks)) {
        if (isset($description))  $description  = (FLG_NR) ? nl2p($description)  : nl2br($description);
        if (isset($mt_text_more)) $mt_text_more = (FLG_NR) ? nl2p($mt_text_more) : nl2br($mt_text_more);
        if (isset($mt_excerpt))   $mt_excerpt   = (FLG_NR) ? nl2p($mt_excerpt)   : nl2br($mt_excerpt);
    }
    //現在時刻を取り出して必要な箇所に適用する。
    $filename = ($postid) ? $postid : date(YEAR_FMT."md_His"); //現在日時から名前生成する。
    $date     = array(date(YEAR_FMT),date("m"),date("d"),date("H"),date("i"),date("s"));
    $datetime = date("Y-m-d H:i:s", time());

    //URLならばリンクを貼る。
    //$description = ereg_replace("(^|[^=\"'])(https?|ftp|news)(://[[:alnum:]\+\$\;\?\.%,!#~*/:@&=_-]+)","\\1<a href=\"\\2\\3\" target=\"_blank\">\\2\\3</a>",$description);
    $newdata  = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n";
    $newdata .= "<rNote xml:lang=\"ja\" xmlns=\"http://rinn.e-site.jp/rnote/\">\n";
    $newdata .= "<Title>" . $title . "</Title>\n";
    $newdata .= "<Keys>" . $mt_keywords . "</Keys>\n";
    if (!$mt_allow_comments) {
        $newdata .= "<WB_Accept>0</WB_Accept>\n"; //$mt_allow_comments が false のときだけ 0 に設定してタグを挿入する。
    }
    $newdata .= "<Date>";
    $newdata .= $datetime;
    $newdata .= "</Date>\n";
    $newdata .= "<Text type=\"text/html\" xml:lang=\"ja\" xmlns=\"http://www.w3.org/1999/xhtml\">\n" . $description . "\n</Text>\n";
    if (strlen($mt_text_more) >= 1) {
        $newdata .= "<".TAG_ADDITIONAL_TEXT.">" . $mt_text_more . "</".TAG_ADDITIONAL_TEXT.">\n";
    }
    if (strlen($mt_excerpt) >= 1) {
        $newdata .= "<Description>" . $mt_excerpt . "</Description>\n";
    }
    $newdata .= "</rNote>\n";

    //書き込むファイルのパスを作る。
    $url = DIR_DATA.$c_key.'/'.$filename.'.xml';
    $lockfilename = 'lock/lock_post';
    $lockfp = @fopen($lockfilename,'w'); //ロックファイルでロックする。
    if (!$lockfp) return FALSE;
    flock($lockfp, LOCK_EX);
    //-------------------------------------------
    $fp = xrFopen($url, "w", 0666);
    if($fp) {
        set_file_buffer($fp, 0);
        flock($fp, LOCK_EX);
        rewind($fp);
        fwrite($fp, $newdata);
        flock($fp, LOCK_UN);
        fclose($fp);
    }
    //-------------------------------------------
    flock($lockfp, LOCK_UN); fclose($lockfp); //ロックファイルを閉じる。
    WriteLog("Out createPost\n");
    if(!$fp) return FALSE;

    //更新処理でかつ元ファイルと新ファイルのキーが異なるときは、元ファイルを削除する。
    if ($postid && ($c_key != $oc_key)) {
        $url = DIR_DATA.$oc_key.'/'.$filename.'.xml';
        if (is_file($url)) xrUnlink($url);
    }
    return $filename;
}

function WriteLog($log)
{
    $lockfilename = 'lock/lock_xmlrpclog';
    $lockfp = @fopen($lockfilename,'w'); //ロックファイルでロックする。
    if (!$lockfp) return FALSE;
    flock($lockfp, LOCK_EX);
    //-------------------------------------------
    $fp = @fopen(XMLRPC_LOG_TXT,"a");
    if ($fp) {
        set_file_buffer($fp, 0);
        flock($fp, LOCK_EX);
        fwrite($fp, $log);
        flock($fp, LOCK_UN);
        fclose($fp);
    }
    //-------------------------------------------
    flock($lockfp, LOCK_UN); fclose($lockfp); //ロックファイルを閉じる。
    if (!$fp) return FALSE;
    return TRUE;
}

function logParams($msg, $params)
{
    $param_num = $params->getNumParams();
    $log = "\n---".date("Y-m-d H:i:s")."---\n".$msg."\n"."number of parameter:[".$param_num."]\n";
    WriteLog($log);
    for ($i = 0; $i < $param_num; $i++) {
        $val = $params->getParam($i);
        $val2 = $val->getval();
        $log = "parameter [".$i."]=[".$val2."]\n";
        WriteLog($log);
    }
}

function get_iso_8601_date($int_date)
{
   //$int_date: current date in UNIX timestamp
   $date_mod = date('Y-m-d\TH:i:s', $int_date);
   $pre_timezone = date('O', $int_date);
   $time_zone = substr($pre_timezone, 0, 3).":".substr($pre_timezone, 3, 2);
   $date_mod .= $time_zone;
   return $date_mod;
}

function GetPostContents($akey, $postid, $userid)
{
    $src = DIR_DATA.$akey.'/'.$postid.'.xml';

    $contents = xrFile_get_contents($src);
    if (!$contents) return FALSE;

    //必要な情報を求める。
    $tg_title     = get_XMLTag(TAG_TITLE, $contents);
    $tg_keys      = get_XMLTag(TAG_KEYS, $contents);
    $tg_date      = get_XMLTag(TAG_DATE, $contents);
    $bAccept      = (get_XMLTag(TAG_WBACCEPT, $contents)==='0') ? 0 : 1;
    $mt_excerpt   = get_XMLTag(TAG_DESCRIPTION, $contents);
    $mt_text_more = get_XMLTag(TAG_ADDITIONAL_TEXT, $contents);
    $contents     = get_XMLTag(TAG_TEXT, $contents);
    //
    $filename = eregi_replace("\.xml", "", basename($src));
    $date = array(strftime("%".YEAR_FMT, strtotime($tg_date)), 
        strftime("%m", strtotime($tg_date)), 
        strftime("%d", strtotime($tg_date)), 
        strftime("%H", strtotime($tg_date)),
        strftime("%M", strtotime($tg_date)),
        strftime("%S", strtotime($tg_date)));
    //構造体を作る。
    //$iso8601date = get_iso_8601_date(strtotime($tg_date));
    $iso8601date = strftime('%Y%m%dT%H:%M:%S', strtotime($tg_date));
    $ret = array('userid'    => new XML_RPC_Value($userid,'string'),//リクエスト者自身のIDを返す。
                 'dateCreated' => new XML_RPC_Value($iso8601date,'dateTime.iso8601'),//ex.20050814T17:57:11Z
                 'postid'    => new XML_RPC_Value($postid,'string'),
                 'description' => new XML_RPC_Value($contents,'string'),
                 'title'     => new XML_RPC_Value($tg_title,'string'),
                 'link'      => new XML_RPC_Value(SITE_URL.FNAME_RNOTE.'?u='.$akey.'/'.$postid.'.htm','string'),
                 'permaLink' => new XML_RPC_Value(SITE_URL.FNAME_RNOTE.'?u='.$akey.'/'.$postid.'.htm','string'),
                 'mt_excerpt' => new XML_RPC_Value($mt_excerpt,'string'),
                 'mt_text_more' => new XML_RPC_Value($mt_text_more,'string'),
                 'mt_allow_comments' => new XML_RPC_Value($bAccept,'int'),
                 'mt_allow_pings' => new XML_RPC_Value($bAccept,'int'),
                 'mt_convert_breaks' => new XML_RPC_Value('none','string'),
                 'mt_keywords' => new XML_RPC_Value($tg_keys,'string')
                );
    return $ret;
}

//getFilePathFromPostid --- post id を元にして、全カテゴリーを検索して当該ファイルを探す。
//
//[戻り値]
//  当該ファイルのパス。見つからなかったら FALSE を返す。
function getFilePathFromPostid($postid)
{
    global $category_key;

    for ($i = 0; $i < count($category_key); $i++) {
        $src = DIR_DATA.$category_key[$i].'/'.$postid.'.xml';
        clearstatcache();
        if (xrFile_exists($src) && is_file($src)) {
            return $src;
        }
    }
    return FALSE;
}

//getCategoryKeyFromPostid --- post id を元にして、全カテゴリーを検索して当該ファイルを探す。
//
//[戻り値]
//  カテゴリーのキー文字列。見つからなかったら FALSE を返す。
function getCategoryKeyFromPostid($postid)
{
    global $category_key;

    for ($i = 0; $i < count($category_key); $i++) {
        $src = DIR_DATA.$category_key[$i].'/'.$postid.'.xml';
        clearstatcache();
        if (xrFile_exists($src) && is_file($src)) {
            return $category_key[$i];
        }
    }
    return FALSE;
}

function getParamByType(&$dst, $param, $type, $name = '')
{
    if (!isset($param)) return FALSE;
    $val = $param->getval();
    if ($param->scalartyp() != 'struct') {
        if ($param->scalartyp() != $type) return FALSE;
        $dst = $val;
    } else {
        $elem = $param->structmem($name);
        if (!isset($elem)) return FALSE;
        if ($elem->scalartyp() != $type) return FALSE;
        $dst = $elem->getval();
    }
    return TRUE;
}

//RebuildCacheAll --- 再構築処理。
function RebuildCacheAll()
{
    $url = SITE_URL.FNAME_RNOTE;
    $ret = send_http($url, "POST", "", array('mode'=>'admin',"PHP_AUTH_USER"=>ADMIN_ID,"PHP_AUTH_PW"=>ADMIN_PASSWORD));
    if(!ereg("終了しました",$ret)) return FALSE;
    return TRUE;
}

//setCategoryFromPostid  --- IDで指定された記事のカテゴリーを設定します。
//[引数]
//    string postid     : カテゴリー設定対象の記事ID。
//    string dst_cate   : 設定するカテゴリーのキー。
//[戻り値]
//    成功:TRUE、失敗:FALSE。
function setCategoryFromPostid($postid, $dst_cate)
{
    global $category_name;

WriteLog("---\nIn setCategoryFromPostid\n");
    //設定先のカテゴリーが実在しないならエラーにする。
    if (!isset($category_name[$dst_cate])) return FALSE;
    // postid を元にして、全カテゴリーを検索して当該ファイルを探す。
    $akey = getCategoryKeyFromPostid($postid);
    if (!$akey) return FALSE;
    //移動元が draft で、移動先が draft 以外なら、 .category を作り移動先カテゴリーキーを入れておく。
    //公開はあくまでも、当該記事に対する publishPost で行われるようにするため。
    if (('draft' == $akey) && ('draft' != $dst_cate)) {
        if (!createCategoryFile($postid, $dst_cate)) return FALSE;
        //if (!createCategoryFile($postid."-test", $dst_cate)) return FALSE;
    } else {
      if (!movePost($postid, $dst_cate, $akey)) return FALSE;
    }
WriteLog("Out setCategoryFromPostid\n");
    return TRUE;
}

//isOptForceConvertBreaks --- オプションのチェック。（強制改行置き換え処理フラグ。）
//[引数]
//    string[] user : ユーザー情報のオプション文字列の連想配列。
//[戻り値]
//    設定されている:boolean TRUE, 設定されていない: FALSE 。
function isOptForceConvertBreaks($user)
{
    $opt = $user['options'];
    $lockfilename = 'lock/lock_strtok';
    $lockfp = @fopen($lockfilename,'w'); //ロックファイルでロックする。
    if (!$lockfp) return xrErrResp(0, XR_ERR_FUNC, "lock");
    flock($lockfp, LOCK_EX);
    //-------------------------------------------
    //画像ファイルなどが保管されるので実行属性付きで作成する。
    // Read Write のみだと file_exist(), fopen() などで失敗する。
    $ret = FALSE;
    $tok = strtok($opt, ",");
    while ($tok) {
        if (!strcmp(mb_strtolower($tok), XR_OPT_FORCE_CONVERT_BREAKS)) {
            $ret = TRUE;
            break;
        }
        $tok = strtok(",");
    }
    //-------------------------------------------
    flock($lockfp, LOCK_UN); fclose($lockfp); //ロックファイルを閉じる。
    return $ret;
}

//isOptForceAllowComments --- オプションのチェック。（強制コメント許可フラグ。）
//[引数]
//    string[] user : ユーザー情報のオプション文字列の連想配列。
//[戻り値]
//    設定されている:boolean TRUE, 設定されていない: FALSE 。
function isOptForceAllowComments($user)
{
    $opt = $user['options'];
    $lockfilename = 'lock/lock_strtok';
    $lockfp = @fopen($lockfilename,'w'); //ロックファイルでロックする。
    if (!$lockfp) return xrErrResp(0, XR_ERR_FUNC, "lock");
    flock($lockfp, LOCK_EX);
    //-------------------------------------------
    //画像ファイルなどが保管されるので実行属性付きで作成する。
    // Read Write のみだと file_exist(), fopen() などで失敗する。
    $ret = FALSE;
    $tok = strtok($opt, " ,");
    while ($tok) {
        if (!strcmp(mb_strtolower($tok), XR_OPT_FORCE_ALLOW_COMMENTS)) {
            $ret = TRUE;
            break;
        }
        $tok = strtok(" ,");
    }
    //-------------------------------------------
    flock($lockfp, LOCK_UN); fclose($lockfp); //ロックファイルを閉じる。
    return $ret;
}

//isOptForcePublish --- オプションのチェック。（newPost()即自動再構築フラグ。）
//[引数]
//    string[] user : ユーザー情報のオプション文字列の連想配列。
//[戻り値]
//    設定されている:boolean TRUE, 設定されていない: FALSE 。
function isOptForcePublish($user)
{
    $opt = $user['options'];
    $lockfilename = 'lock/lock_strtok';
    $lockfp = @fopen($lockfilename,'w'); //ロックファイルでロックする。
    if (!$lockfp) return xrErrResp(0, XR_ERR_FUNC, "lock");
    flock($lockfp, LOCK_EX);
    //-------------------------------------------
    //画像ファイルなどが保管されるので実行属性付きで作成する。
    // Read Write のみだと file_exist(), fopen() などで失敗する。
    $ret = FALSE;
    $tok = strtok($opt, " ,");
    while ($tok) {
        if (!strcmp(mb_strtolower($tok), XR_OPT_FORCE_PUBLISH)) {
            $ret = TRUE;
            break;
        }
        $tok = strtok(" ,");
    }
    //-------------------------------------------
    flock($lockfp, LOCK_UN); fclose($lockfp); //ロックファイルを閉じる。
    return $ret;
}

//newPost --- 新しい記事を投稿します。公開することもできます。
//[引数]
//    string  blogid   : blog ID。
//    string  username : blog のユーザ名(アカウント)。
//    string  password : パスワード。
//    struct  content  : content 構造体。
//    boolean publish  : 下書きか公開か。（0: 下書き。1: 公開。）
//[注記] content 構造体の要素は以下のとおり：
//    string   title       :  件名。
//    string   description :  本文。
//    ISO.8601 dateCreated :  投稿日時。（ISO.8601 形式。タイムゾーン指定不可。省略可能。）
//  ※Movable Type 拡張オプションとして以下の7項目を受け取る可能性がある。（省略可能。）
//    int      mt_allow_comments : コメントを受け入れるか否か。
//    int      mt_allow_pings    : トラックバックPingを受け入れるか否か。
//    string   mt_convert_breaks : 本文の改行の扱い。
//                                 0/none : 本文および追記中の改行を <br /> or </p><p> に変換しない。
//                                 1/convert_breaks : 本文および追記中の改行を <br /> or </p><p> に変換する。
//    string   mt_text_more      : 追記。
//    string   mt_excerpt        : 概要。
//    string   mt_keywords       : キーワード。（カンマ区切りで複数可能。）
//    array    mt_tb_ping_urls   : Trackback ping を送信する宛先 URL リスト。
//[戻り値]
//    成功:string postid （投稿した「記事ID」。）失敗: fault 。
function newPost($params)
{
    global $xmlrpcerruser; // import user errcode value

    WriteLog("---\nIn newPost:\n");
    for ($i = 0; $i < $params->getNumParams(); $i++) {
        $val = $params->getParam($i);
        switch ($i) {
        case 0: //blog ID
            // BlogPet が int で送ってくるそうなので対応した。( RFC では string )
            if (!getParamByType($blogid, $val, 'string')) {
                if (!getParamByType($blogid, $val, 'int')) {
                    return xrErrResp(0, XR_ERR_ARG, "blogid");
                }
            }
            $blogid = $val->getval();
            break;
        case 1: //blog のユーザーアカウント
            if (!getParamByType($username, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "username");
            break;
        case 2: //上記アカウントのログインパスワード
            if (!getParamByType($password, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "password");
            break;
        case 3: //content構造体
            if (!getParamByType($title, $val, 'string', 'title')) return xrErrResp(0, XR_ERR_ARG, "title");
            if (!getParamByType($description, $val, 'string', 'description')) ;//無ければ何もセットしない。
            if (!getParamByType($mt_allow_comments, $val, 'int', 'mt_allow_comments')) ;
            if (!getParamByType($mt_allow_pings, $val, 'int', 'mt_allow_pings')) ;
            if (!getParamByType($mt_convert_breaks, $val, 'string', 'mt_convert_breaks')) $mt_convert_breaks = 'none';
            if (!getParamByType($mt_text_more, $val, 'string', 'mt_text_more')) ;//無ければ何もセットしない。
            if (!getParamByType($mt_excerpt, $val, 'string', 'mt_excerpt')) ;//無ければ何もセットしない。
            if (!getParamByType($mt_keywords, $val, 'string', 'mt_keywords')) $mt_keywords = '';
            if (!getParamByType($mt_tb_ping_urls, $val, 'array', 'mt_tb_ping_urls')) ;
            break;
        case 4: //公開するか否か
            if (!getParamByType($publish, $val, 'boolean')) {
                //moblog.uva.ne.jp が int で値を返してくるため。
                if (!getParamByType($publish, $val, 'int')) {
                    return xrErrResp(0, XR_ERR_ARG, "publish");
                }
            }
            break;
        }
    }
    //許可されたユーザーかチェックする。
    $user = userCheck($username,$password);
    if ((SITE_BLOGID != $blogid) || !isset($user)) return xrErrResp(0, XR_ERR_USER, "");
    //強制改行置き換えフラグのチェック。
    if (isOptForceConvertBreaks($user)) {
        $mt_convert_breaks = 'convert_breaks';
    }
    //強制コメント許可フラグのチェック。
    if (isOptForceAllowComments($user)) {
        $mt_allow_comments = 1;
    }
    //編集した記事は draft に移動する。その後 publish により公開される。
    //投稿に成功したら、記事のIDを返す。
    $postid = createPost(0, DRAFT_KEY, $mt_convert_breaks, $title, $description, $mt_text_more, $mt_allow_comments, $mt_allow_pings, $mt_excerpt, $mt_keywords);
    if (!$postid) return xrErrResp(0, XR_ERR_FUNCTION, "createPost");
    // BlogPet からの書き込み専用の処理。 newPost しかしてこないので即時再構築。
    //if ((BLOGPET_ID == $username) && strstr($title,'(BlogPet)')) {
    if (isOptForcePublish($user)) {
        //rnote_config.php のユーザー情報 ( $XR-USER ) にある primaricategory をキーとする
        //カテゴリーに、自動的にカテゴリーを設定して投稿する。
        setCategoryFromPostid($postid, $user['primarycategory']);
        publishPost($postid);
    }
    WriteLog("mt_tb_ping_urls=[".$mt_tb_ping_urls."]\n");
    //Trackback ping URL のリストを一時ファイルに作成する。
    if (isset($mt_tb_ping_urls) && count($mt_tb_ping_urls)) {
        WriteLog("call createTBurlFile()\n");
        createTBurlFile($postid, $title, $mt_excerpt, $mt_tb_ping_urls);
    }
    $val = new XML_RPC_Value($postid, 'string'); 
    WriteLog("out newPost\n");
    return xrResp($val);
}

//getPost --- IDで指定された記事を取得します。
//[引数]
//    string postid   : 取得したい記事の記事ID 
//    string username : blog のユーザ名(アカウント) 
//    string password : 上記ユーザの blog への login 用パスワード 
//[戻り値]
//  成功:以下の要素を含む構造体。
//    string   userid            : 投稿者のユーザID。
//    ISO.8601 dateCreated       : 記事の投稿日時。
//    string   postid            : 記事ID。
//    string   description       : 本文。
//    string   title             : 件名。
//    string   link              : 記事 の URL (下記の permalink と同一)。
//    string   permaLink         : permalink (ココログでは「固定リンク」)。
//  ※以下は Movable Type での追加拡張オプション。
//    string   mt_excerpt        : 概要。
//    string   mt_text_more      : 追記。
//    int      mt_allow_comments : コメントを受け入れるか。
//    int      mt_allow_pings    : トラックバック Ping を受け入れるか。
//    string   mt_convert_breaks : 改行の扱いをどうするか。
//    string   mt_keywords;      : キーワード。
//  失敗: fault 。
//[注意]
//  link と permaLink は、どちらも 蓄積されている記事への URL 。
function getPost($params)
{
    global $xmlrpcerruser; // import user errcode value
    global $category_key, $category_name;

    for ($i = 0; $i < $params->getNumParams(); $i++) {
        $val = $params->getParam($i);
        switch ($i) {
        case 0: //postid
            if (!getParamByType($postid, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "postid");
            break;
        case 1: //blog のユーザーアカウント
            if (!getParamByType($username, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "username");
            break;
        case 2: //上記アカウントのログインパスワード
            if (!getParamByType($password, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "password");
            break;
        }
    }
    //許可されたユーザーかチェックする。
    if (!userCheck($username,$password)) return xrErrResp(0, XR_ERR_USER, "");

    $akey = getCategoryKeyFromPostid($postid);
    if (!$akey) return xrErrResp(0, XR_ERR_POSTID, "");
    $ret = GetPostContents($akey, $postid, $userid);

    $val = new XML_RPC_Value($ret, "struct"); 
    return xrResp($val);
}

//metaWeblog_editPost --- IDで指定された記事を上書き更新します。
//[引数]
//    string  postid,   : 上書き対象の記事ID。
//    string  username, : blog のユーザ名(アカウント)。
//    string  password, : パスワード。
//    struct  content,  : content構造体。
//    boolean publish   : 公開するか（ 0: 下書きとして投稿。1: 公開して登録。）
//    注意: content 構造体は、次のような標準的な項目を含みます:
//    string   title,      : 件名。
//    string   description : 本文。
//    ISO.8601 dateCreated : 投稿日時。(省略すると編集前の日時を引き継ぐ。)
//  ※Movable Type 用の拡張オプションとして以下を受信する可能性がある:
//    int      mt_allow_comments : コメントを受け入れるか。
//    int      mt_allow_pings    : トラックバックPingを受け入れるか。
//    string   mt_convert_breaks : 本文の改行の扱いをどうするか。
//    string   mt_text_more      : 追記。 
//    string   mt_excerpt        : 概要。
//    string   mt_keywords       : キーワード。
//    array    mt_tb_ping_urls   : Trackback ping 宛先 URL リスト。
//[戻り値]
//    成功: true 。失敗: fault 。
function metaWeblog_editPost($params)
{
    global $xmlrpcerruser; // import user errcode value
    global $category_key, $category_name;

    WriteLog("---\nIn editPost:\n");
    $has_content = FALSE;//content構造体に値がセットされていれば TRUE となる。
    //値がセットされていなかったとき用の値。
    $mt_convert_breaks = 'none';
    $mt_keywords = '';
    for ($i = 0; $i < $params->getNumParams(); $i++) {
        $val = $params->getParam($i);
        switch ($i) {
        case 0: //blog ID
            if (!getParamByType($postid, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "postid");
            break;
        case 1: //blog のユーザーアカウント
            if (!getParamByType($username, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "username");
            break;
        case 2: //上記アカウントのログインパスワード
            if (!getParamByType($password, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "password");
            break;
        case 3: //content構造体
            if (getParamByType($title, $val, 'string', 'title')) $has_content = TRUE;
            if (getParamByType($description, $val, 'string', 'description')) $has_content = TRUE;
            if (getParamByType($mt_allow_comments, $val, 'int', 'mt_allow_comments')) $has_content = TRUE;
            if (getParamByType($mt_allow_pings, $val, 'int', 'mt_allow_pings')) $has_content = TRUE;
            if (getParamByType($mt_convert_breaks, $val, 'string', 'mt_convert_breaks')) $has_content = TRUE;
            if (getParamByType($mt_text_more, $val, 'string', 'mt_text_more')) $has_content = TRUE;
            if (getParamByType($mt_excerpt, $val, 'string', 'mt_excerpt')) $has_content = TRUE;
            if (getParamByType($mt_keywords, $val, 'string', 'mt_keywords')) $has_content = TRUE;
            if (getParamByType($mt_tb_ping_urls, $val, 'array', 'mt_tb_ping_urls')) $has_content = TRUE;
            break;
        case 4: //公開するか否か
            if (!getParamByType($publish, $val, 'boolean')) {
                if (!getParamByType($publish, $val, 'int')) {
                    //moblog.uva.ne.jp が値を int で返してくるため。
                    return xrErrResp(0, XR_ERR_ARG, "publish");
                }
            }
            break;
        }
    }
    //許可されたユーザーかチェックする。
    if (!userCheck($username,$password)) return xrErrResp(0, XR_ERR_USER, "");
    //content構造体に何がしかの値がセットされていれば createPost() を実行する。
    if ($has_content) {
        //投稿に成功したら、記事のIDを返す。
        if (!createPost($postid, DRAFT_KEY, $mt_convert_breaks, $title, $description, $mt_text_more, $mt_allow_comments, $mt_allow_pings, $mt_excerpt, $mt_keywords)) {
            return xrErrResp(0, XR_ERR_FUNC, "post.");
        }
    }
    //moblog.uva.ne.jp の記事公開の仕組みに対応するためこのようにする。
    //moblog.uva.ne.jp では newPost(publish=0) -> setPostCategories(publish=0) -> editPost(publish=1) として
    //記事を公開する流れを実現しているので。（なぜ newPost -> setPostCategories -> mt.publishPost としないのだろうか。）
    //publish が真ならば再構築を実行する。
    if ($publish) {
        publishPost($postid);
    }
    //Trackback ping URL のリストを一時ファイルに作成する。
    if (isset($mt_tb_ping_urls) && count($mt_tb_ping_urls)) {
        createTBurlFile($postid, $title, $mt_excerpt, $mt_tb_ping_urls);
    }
    $val = new XML_RPC_Value(true, 'boolean'); 
    WriteLog("---\nOut editPost:\n");
    return xrResp($val);
}

//getRecentPosts --- 最新の記事をシステムから取ってきます。
//[引数]
//    string blogid        : blog ID。
//    string username      : blog のユーザ名(アカウント)。
//    string password      : 上記ユーザの blog への login 用パスワード。
//    int    numberOfPosts : 取得する記事数（0なら全件 ubicast Blogger の拡張）。
//[戻り値]
//    成功: 以下の要素を含む構造体の記事数分の配列。
//    ISO.8601 dateCreated       : 投稿日時。
//    string   userid            : 投稿者のユーザID。
//    string   postid            : 記事ID。
//    string   description       : 本文。
//    string   title             : 件名。
//    string   link              : 記事への URL。
//    string   permaLink         : 記事への URL。（ link と同一。）
//  ※以下は Movable Type の拡張オプション。
//    string   mt_excerpt        : 概要。
//    string   mt_text_more      : 追記。
//    int      mt_allow_comments : コメントを受け入れるか。
//    int      mt_allow_pings    : トラックバック Ping を受け入れるか。
//    string   mt_convert_breaks : 改行の扱い。
//    string   mt_keywords       : キーワード。
//    失敗: fault 。
//[備考]
//    本文＆追記を CDATA で返してみたが BlogWrite, ubicast Blogger 共に閉じ括弧
//  ']]>' が本文末尾に表示される不都合が出たためとりやめた。
function getRecentPosts($params)
{
    global $xmlrpcerruser; // import user errcode value
    global $category_key, $category_name;

    $numberOfPosts = 15;

    for ($i = 0; $i < $params->getNumParams(); $i++) {
        $val = $params->getParam($i);
        switch ($i) {
        case 0: //blogid
            // BlogPet が int で送ってくるそうなので対応した。( RFC では string )
            if (!getParamByType($blogid, $val, 'string')) {
                if (!getParamByType($blogid, $val, 'int')) return xrErrResp(0, XR_ERR_ARG, "blogid");
            }
            $blogid = $val->getval();
            break;
        case 1: //blog のユーザーアカウント
            if (!getParamByType($username, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "username");
            break;
        case 2: //上記アカウントのログインパスワード
            if (!getParamByType($password, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "password");
            break;
        case 3: //取得する記事の数。
            if (!getParamByType($numberOfPosts, $val, 'int')) return xrErrResp(0, XR_ERR_ARG, "numberOfPosts");
            break;
        }
    }
    //許可されたユーザーかチェックする。
    if (!userCheck($username,$password)) return xrErrResp(0, XR_ERR_USER, "");

    $ret = array();
    //まだエントリが一つも無い。
    if(!xrFile_exists(FNAME_FILECACHE)){
        //空の配列を返す。
        $val = new XML_RPC_Value($ret, "array"); 
        return xrResp($val);
    }

//-----
    $fl = file(FNAME_FILECACHE);
    unset($filelist);
    //unset($ret);
    $n = 0;
    foreach($fl as $a){
        list($path,$date,$title,$wbaccept) = explode("\t",str_replace("\n",'',$a));
        if(!xrFile_exists(DIR_DATA.$path)){
            WriteLog("getRecentPosts: xrFile_exists: not exist[".DIR_DATA.$path."]");
            //指定されたファイルがなかった。再構築が必要。
            if (!RebuildCacheAll()) return xrErrResp(0, XR_ERR_FUNC, "send_http");
            //再構築してもダメならエラー
            if(!xrFile_exists(DIR_DATA.$path)) return xrErrResp(0, XR_ERR_FUNC, "file_exists");
        }
        $buf = xrFile_get_contents(DIR_DATA.$path);
        SplitPath($path,$dir,$fname);
        $fname = basename($fname,".xml");
        SplitPath($_SERVER['PHP_SELF'],$base,$tmp);
        if(!USE_MOD_REWRITE) $base .= FNAME_RNOTE.'?u=';
        $url = 'http://'.$_SERVER["HTTP_HOST"].$base.ext_change($path,EXT_ACCESS);
        // コメントについては、記事で明示的に拒否してない時以外は true とした。
        // もしかしたら受付期間過ぎてるかもだけど、真面目にやるとちょっと面倒だしー
        $bAccept = (get_XMLTag(TAG_WBACCEPT,$buf)==='0') ? 0 : 1;
        $text = replaceRightBlacket(get_XMLTag(TAG_TEXT,$buf));
        $add = replaceRightBlacket(get_XMLTag(TAG_ADDITIONAL_TEXT,$buf));
        $mt_excerpt = replaceRightBlacket(GetDescription($buf,128));
        $mt_keywords = replaceRightBlacket(get_XMLTag(TAG_KEYS,$buf));
        //dateCreated は当該 weblog が設置されているサーバのある場所の現地時間。
        //UTC だと BlogWrite, ubicast blogger 共に9時間前の時間として読み込んでしまうのでローカル時間にした。
        //$iso8601date = gmdate("Y-m-d\TH:i:s",$date).'Z';
        $iso8601date = strftime('%Y%m%dT%H:%M:%S', $date);
        array_push($ret,
            new XML_RPC_Value(
                array(
                'dateCreated' => new XML_RPC_Value($iso8601date, 'dateTime.iso8601'), 
                'userid' => new XML_RPC_Value(PET_ID, 'string'),                         // 投稿者のユーザID
                // どうするか一考の余地あり。とりあえずファイル名で、別DIRに同名ファイルが無い事を祈る！
                'postid' => new XML_RPC_Value($fname, 'string'),                         // 記事ID
                //'description' => new XML_RPC_Value("<![CDATA[".$text."]]>", 'string'), // 記事本文
                'description' => new XML_RPC_Value($text, 'string'),                     // 記事本文
                'title' => new XML_RPC_Value($title, 'string'),                          // 記事の件名
                'link' => new XML_RPC_Value($url, 'string'),                             // url,permaLink
                'permaLink' => new XML_RPC_Value($url, 'string'),                        // url,permaLink
                'mt_excerpt' => new XML_RPC_Value($mt_excerpt, 'string'),                // 概要
                //'mt_text_more' => new XML_RPC_Value("<![CDATA[".$add."]]>", 'string'), // 追記
                'mt_text_more' => new XML_RPC_Value($add, 'string'),                     // 追記
                'mt_allow_comments' => new XML_RPC_Value($bAccept, 'int'),               // コメントを受け入れるか否か
                'mt_allow_pings' => new XML_RPC_Value($bAccept, 'int'),                  // トラックバック Ping を受け入れるか否か
                // none:タグを変換せずそのまま返してます。
                'mt_convert_breaks' => new XML_RPC_Value('none', 'string'),              // 改行の扱いをどうするか
                'mt_keywords' => new XML_RPC_Value($mt_keywords, 'string')), "struct")); // キーワード（属性キーをセット。）
        // 指定数取得した？
        //if ($numberOfPosts) {
            if((++$n) >= $numberOfPosts) break;
        //}
    }
//-----
    $val = new XML_RPC_Value($ret, "array");
    return xrResp($val);
}

//mt_supportedMethods --- 投稿API で使用可能な手続きの配列を返す。
//[引数]
//    string blogid        : blog ID。
//    string username      : blog のユーザ名(アカウント)。
//    string password      : 上記ユーザの blog への login 用パスワード。
//[戻り値]
//  成功: 使用可能な手続き名の配列を返す。失敗: fault 。
function mt_supportedMethods($params)
{
    global $xmlrpcerruser; // import user errcode value

    for ($i = 0; $i < $params->getNumParams(); $i++) {
        $val = $params->getParam($i);
        switch ($i) {
        case 0: //blog ID
            // BlogPet が int で送ってくるそうなので対応した。( RFC では string )
            if (!getParamByType($blogid, $val, 'string')) {
                if (!getParamByType($blogid, $val, 'int')) return xrErrResp(0, XR_ERR_ARG, "blogid");
            }
            $blogid = $val->getval();
            break;
        case 1: //blog のユーザーアカウント
            if (!getParamByType($username, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "username");
            break;
        case 2: //上記アカウントのログインパスワード
            if (!getParamByType($password, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "password");
            break;
        }
    }
    //許可されたユーザーかチェックする。
    if (!userCheck($username,$password)) return xrErrResp(0, XR_ERR_USER, "");
    //サポートしている関数の一覧を返す。自分自身は含まない。
    $ret = array(
            new XML_RPC_Value("blogger.deletePost", 'string'),
            new XML_RPC_Value("blogger.getUserInfo", 'string'),
            new XML_RPC_Value("blogger.getUsersBlogs", 'string'),
            new XML_RPC_Value("mt.publishPost", 'string'),
            new XML_RPC_Value("mt.getCategoryList", 'string'),
            new XML_RPC_Value("mt.getPostCategories", 'string'),
            new XML_RPC_Value("mt.setPostCategories", 'string'),
            new XML_RPC_Value("metaWeblog.getCategories", 'string'),
            new XML_RPC_Value("metaWeblog.getPost", 'string'),
            new XML_RPC_Value("metaWeblog.getRecentPosts", 'string'),
            new XML_RPC_Value("metaWeblog.newMediaObject", 'string'),
            new XML_RPC_Value("metaWeblog.editPost", 'string'),
            new XML_RPC_Value("metaWeblog.newPost", 'string')
          );
    $val = new XML_RPC_Value($ret, "array"); 
    return xrResp($val);
}

function publishPost($postid)
{
    // .category を読んでみて、移動予定のカテゴリーがセットされていればそちらに移動する。
    // 移動に成功したとき、元のディレクトリにあった .category ファイルは削除される。
    $dst_cate = readCategoryFile($postid);
    if ($dst_cate) {
        if (!movePost($postid, $dst_cate, 'draft')) {
            return FALSE;
        }
    }

    //Trackback ping 送信先リストファイルがあれば、送信する。
    $title = '';
    $excerpt = '';
    $tb_urls = array();
    if (readTBurlFile($postid, $title, $excerpt, $tb_urls)) {
        //なぜか送信より先に削除しないと「削除関数自体が呼ばれない」ので先に削除する。
        removeTBurlFile($postid);
        sendTrackBackPing($postid, $title, $excerpt, $tb_urls);
    }

    //再構築処理。
    if (!RebuildCacheAll()) {
        return FALSE;
    }
    return TRUE;
}

//mt_publishPost --- 指定記事の公開および再構築処理。
//    1. setPostCategories() でつくられた .category を参照し entries/draft から
//       当該投稿記事の .xml ファイルを移動する。
//    2. 再構築処理を呼び出す。（既に公開されていても再構築を呼ぶ。）
//[引数]
//    string postid   : 公開対象の記事ID。
//    string username : blog のユーザ名(アカウント)。
//    string password : パスワード。
//[戻り値]
//    成功: true 。失敗: fault 。
//[注意]
//    記事の状態が更新されても、 本 API は Update ping の送出を行わない。
function mt_publishPost($params)
{
    global $xmlrpcerruser; // import user errcode value
    global $category_key, $category_name;

WriteLog("---\nIn publishPost:\n");
    for ($i = 0; $i < $params->getNumParams(); $i++) {
        $val = $params->getParam($i);
        switch ($i) {
        case 0: //postid
            if (!getParamByType($postid, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "postid");
            break;
        case 1: //blog のユーザーアカウント
            if (!getParamByType($username, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "username");
            break;
        case 2: //上記アカウントのログインパスワード
            if (!getParamByType($password, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "password");
            break;
        }
    }

    //許可されたユーザーかチェックする。
    if (!userCheck($username,$password)) return xrErrResp(0, XR_ERR_USER, "");
    if (!publishPost($postid)) return xrErrResp(0, XR_ERR_FUNCTION, "publishPost");
WriteLog("Out publishPost:\n");
    $val = new XML_RPC_Value(true, 'boolean'); 
    return xrResp($val);
}

//blogger_deletePost --- 記事を削除します。
//[引数]
//    string  appkey   : 任意の文字列でよい。
//    string  postid   : 削除対象の記事ID。
//    string  username : blog のユーザ名(アカウント)。
//    string  password : 上記ユーザの blog への login 用パスワード。
//    boolean publish  : 0 または 1 を指定(無視される)。
//[戻り値]
//    成功: true，失敗: fault 。
function blogger_deletePost($params)
{
    global $xmlrpcerruser; // import user errcode value
    global $category_key, $category_name;

    for ($i = 0; $i < $params->getNumParams(); $i++) {
        $val = $params->getParam($i);
        switch ($i) {
        case 0: //appkey(何でも良い。)
            if (!getParamByType($appkey, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "appkey");
            break;
        case 1: //postid
            if (!getParamByType($postid, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "postid");
            break;
        case 2: //blog のユーザーアカウント
            if (!getParamByType($username, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "username");
            break;
        case 3: //上記アカウントのログインパスワード
            if (!getParamByType($password, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "password");
            break;
        case 4: //publish（無視してよい。）
            if (!isset($val) || ($val->scalartyp() != 'boolean')) {
                if (!isset($val) || ($val->scalartyp() != 'int')) {
                    //moblog.uva.ne.jp が値を int で返してくるため。
                    return xrErrResp(0, XR_ERR_ARG, "publish");
                }
            }
            $publish = $val->getval();
            break;
        }
    }

    //許可されたユーザーかチェックする。
    if (!userCheck($username,$password)) return xrErrResp(0, XR_ERR_USER, "");

    $akey = getCategoryKeyFromPostid($postid);
    if ($akey) {
        $src = DIR_DATA.$akey.'/'.$postid.'.xml';
        //ファイルを削除する。
        xrUnlink($src);
    } else {
        return xrErrResp(0, XR_ERR_POSTID, "");
    }

    $ret = true;
    $val = new XML_RPC_Value($ret, 'boolean'); 
    return xrResp($val);
}

//getCategoryList --- 指定 blog の全カテゴリーの情報の配列を返す。
//[引数]
//    string blogid   : blog ID。
//    string username : blog のユーザ名(アカウント)。
//    string password : パスワード。
//[戻り値]
//    成功: カテゴリーの数だけ、次の要素を含む構造体の配列を返す。
//    string categoryId   : カテゴリID。
//    string categoryName : カテゴリ名。
//    失敗: fault 。
function getCategoryList($params, $mode)
{
    global $xmlrpcerruser; // import user errcode value
    global $category_key, $category_name;

    for ($i = 0; $i < $params->getNumParams(); $i++) {
        $val = $params->getParam($i);
        switch ($i) {
        case 0: //blogid
            // BlogPet が int で送ってくるそうなので対応した。( RFC では string )
            if (!getParamByType($blogid, $val, 'string')) {
                if (!getParamByType($blogid, $val, 'int')) return xrErrResp(0, XR_ERR_ARG, "blogid");
            }
            $blogid = $val->getval();
            break;
        case 1: //blog のユーザーアカウント
            if (!getParamByType($username, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "username");
            break;
        case 2: //上記アカウントのログインパスワード
            if (!getParamByType($password, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "password");
            break;
        }
    }

    //許可されたユーザーかチェックする。
    if ((SITE_BLOGID != $blogid) || !userCheck($username,$password)) return xrErrResp(0, XR_ERR_USER, "");

    $ret = array();
    if (!strcmp($mode,'mt')) {
        for ($i = 0; $i < count($category_key); $i++) {
            array_push($ret,
                new XML_RPC_Value(
                    array('categoryId' => new XML_RPC_Value($category_key[$i], 'string'), 
                          'categoryName' => new XML_RPC_Value($category_name[$category_key[$i]], 'string')), "struct"));
        }
        $val = new XML_RPC_Value($ret, "array"); 
    } else {
        //metaWeblog API
        //RFC では返す値の名前しか記されていないので曖昧。
        //RFC では struct の scruct を返すとあるが変。 struct の array を返す implementation が実績があるようだ。
        for ($i = 0; $i < count($category_key); $i++) {
            $htmlUrl = SITE_URL.FNAME_RNOTE.'?u='.$category_key[$i];
            $rssUrl = SITE_URL.'rss.rdf';//カテゴリーごとに rss があるという前提か？
            array_push($ret,
                new XML_RPC_Value(
                    array('title' => new XML_RPC_Value($category_name[$category_key[$i]], 'string'), 
                          'categoryid' => new XML_RPC_Value($category_key[$i], 'string'), 
                          'description' => new XML_RPC_Value($category_name[$category_key[$i]], 'string'),
                          'htmlUrl' => new XML_RPC_Value($htmlUrl, 'string'),
                          'rssUrl' => new XML_RPC_Value($rssUrl, 'string')), "struct"));
        }
        $val = new XML_RPC_Value($ret, "array"); 
    }
    return xrResp($val);
}

//metaWeblog_getCategories --- 指定 blog の全カテゴリーの情報の配列を返す。
//[引数]
//    string blogid   : blog ID。
//    string username : blog のユーザ名(アカウント)。
//    string password : パスワード。
//[戻り値]
//    成功すると、カテゴリーの数だけ、次の要素を含む構造体の配列を返す:
//    string desctiption : カテゴリID。
//    string htmlUrl     : カテゴリの名前。
//    string rssUrl      : カテゴリID。
//    string title       : カテゴリ名。
//    string categoryid  : カテゴリ名。
//    なお、構造体のメンバーの出現順序は不定です。
//    失敗すると、fault が返ります。 
//[注意]
//   戻り値の正確な内容が不明のため未チェック。当処理を使用する投稿クライアントは
//   ect だけであり、 Windows 版 ect の品質の問題（挙動不審）のためテストできず。
function metaWeblog_getCategories($params)
{
    return getCategoryList($params, 'wl');
}

//mt_getCategoryList --- 指定 blog の全カテゴリーの情報の配列を返す。
//[引数]
//    string blogid   : blog ID。
//    string username : blog のユーザ名(アカウント)。
//    string password : パスワード。
//[戻り値]
//    成功: カテゴリーの数だけ、次の要素を含む構造体の配列を返す:
//    string categoryId   : カテゴリID。
//    string categoryName : カテゴリ名。
//    失敗: fault 。 
function mt_getCategoryList($params)
{
    return getCategoryList($params, 'mt');
}

//newMediaObject --- 任意のファイルをサーバ上に送信する。
//[引数]
//    string blogid   : blog ID。
//    string username : blog のユーザ名(アカウント)。
//    string password : 上記ユーザの blog への login 用パスワード。
//    struct file     : 送信ファイルの情報。
//      base64 bits : ファイルの中身。（BASE64 で符号化。）
//      string name : ファイル名。
//      string type : ファイル種別 ( ex. 'image/jpeg' )。
//    ※ type [ファイルの種類] は、現状無視されている。
//[戻り値]
//    転送したファイルへの URL を要素とする構造体を返す。（MTの結果を基準にした。）
//      string url; //唯一のメンバー。
function newMediaObject($params)
{
    global $xmlrpcerruser; // import user errcode value
    global $category_key, $category_name;

    clearstatcache();
    WriteLog("---\nIn newMediaObject: \n");
    for ($i = 0; $i < $params->getNumParams(); $i++) {
        $val = $params->getParam($i);
        switch ($i) {
        case 0: //blog ID
            // BlogPet が int で送ってくるそうなので対応した。( RFC では string )
            if (!getParamByType($blogid, $val, 'string')) {
                if (!getParamByType($blogid, $val, 'int')) return xrErrResp(0, XR_ERR_ARG, "blogid");
            }
            $blogid = $val->getval();
            break;
        case 1: //blog のユーザーアカウント
            if (!getParamByType($username, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "username");
            break;
        case 2: //上記アカウントのログインパスワード
            if (!getParamByType($password, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "password");
            break;
        case 3: //ファイル構造体
            if (!isset($val) || ($val->scalartyp() != "struct")) {
                return xrErrResp(0, XR_ERR_ARG, "file");
            }
            $elem = $val->structmem("name");
            if (!isset($elem) || ($elem->scalartyp() != 'string')) {
                return xrErrResp(0, XR_ERR_ARG, "name");
            }
            $name = $elem->getval();
            $elem = $val->structmem("bits");
            //$bits = $elem->getval();
            //$contents = base64_decode($bits);
            if (!isset($elem)) {
                return xrErrResp(0, XR_ERR_ARG, "bits");
            }
            $contents = $elem->getval();// getval() した時点でどうやら decode されているらしい。
            $elem = $val->structmem("type");
            //if (!isset($elem) || ($elem->scalartyp() != 'string')) {
            //    return xrErrResp(0, XR_ERR_ARG, "type");
            //}
            $type = $elem->getval();
            break;
        }
    }

    //許可されたユーザーかチェックする。
    if (!userCheck($username,$password)) return xrErrResp(0, XR_ERR_USER, "");

    //このファイルの実行されているディレクトリからの相対ディレクトリとする。
    //$dst = SITE_DIR.$name;
    //$dst = './'.$name;
    $dst = $name;
    //SITE_URL の末尾には '/' が有り $name の先頭には '/' が無い、という前提。
    $url = SITE_URL.$name;

    SplitPath($dst, $dir, $fname);
    //ディレクトリが存在しなければ作る。
    clearstatcache();
    if (!is_dir($dir)) {
        xrMkdir($dir);
        //画像ファイルなどが保管されるので実行属性付きで作成する。
        // Read Write のみだと file_exist(), fopen() などで失敗する。
        xrChmod($dir, 0777);
    }
    //バイナリファイルを書き込む。
    $lockfilename = 'lock/lock_newmedia';
    $lockfp = @fopen($lockfilename,'w'); //ロックファイルでロックする。
    if (!$lockfp) return xrErrResp(0, XR_ERR_FUNC, "lock");
    flock($lockfp, LOCK_EX);
    //-------------------------------------------
    //画像ファイルなどが保管されるので実行属性付きで作成する。
    // Read Write のみだと file_exist(), fopen() などで失敗する。
    $fp = xrFopen($dst, "wb", 0777);
    if ($fp) {
        set_file_buffer($fp, 0);
        flock($fp, LOCK_EX);
        rewind($fp);
        fwrite($fp, $contents, strlen($contents));
        WriteLog("content length=[".strlen($contents)."]\n");
        flock($fp, LOCK_UN);
        fclose($fp);
    } else {
        WriteLog("returned fp is false\n");
    }
    //-------------------------------------------
    flock($lockfp, LOCK_UN); fclose($lockfp); //ロックファイルを閉じる。
    if (!$fp) xrErrResp(0, XR_ERR_FUNC, "fwrite");

    WriteLog("Out newMediaObject:\n");

    //現状では記事のカテゴリーは一つしかないので、ファイルの存在したカテゴリー１つを返してやる。
    $ret = new XML_RPC_Value(array('url' => new XML_RPC_Value($url, 'string')), "struct");
    return xrResp($ret);
}

//getPostCategories  --- 指定記事に設定されているカテゴリーを取得する。
//[引数]
//    string postid   : カテゴリ取得対象の記事ID。
//    string username : blog のユーザ名(アカウント)。
//    string password : パスワード。
//[戻り値]
//    成功: 次の要素を含む構造体の配列を返す。
//    string  categoryId   : カテゴリID。
//    string  categoryName : カテゴリ名。
//    boolean isPrimary    : プライマリーカテゴリか。
//    失敗: fault 。
function getPostCategories($params)
{
    global $xmlrpcerruser; // import user errcode value
    global $category_key, $category_name;

    for ($i = 0; $i < $params->getNumParams(); $i++) {
        $val = $params->getParam($i);
        switch ($i) {
        case 0: //postid
            if (!getParamByType($postid, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "postid");
            break;
        case 1: //blog のユーザーアカウント
            if (!getParamByType($username, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "username");
            break;
        case 2: //上記アカウントのログインパスワード
            if (!getParamByType($password, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "password");
            break;
        }
    }

    //許可されたユーザーかチェックする。
    if (!userCheck($username,$password)) return xrErrResp(0, XR_ERR_USER, "");

    // post id を元にして、全カテゴリーを検索して当該ファイルを探す。
    $new_keys = array();
    $akey = getCategoryKeyFromPostid($postid);
    if ($akey) {
        // .category を読んでみて、移動予定のカテゴリーがセットされていればそちらを返すようにする。
        $dst_cate = readCategoryFile($postid);
        if ($dst_cate) $akey = $dst_cate;
        array_push($new_keys, $akey);//配列の形で複数渡せるようにしておく。
    }

    //現状では記事のカテゴリーは一つしかないので、ファイルの存在したカテゴリー１つを返してやる。
    $ret = array();
    for ($i = 0; $i < count($new_keys); $i++) {
        //プライマリーカテゴリーは、１件目のカテゴリーに固定しておく。（現在カテゴリーは１つしかない。）
        if ($i == 0) $isPrimary = true; else $isPrimary = false;
        array_push($ret,
            new XML_RPC_Value(
                array('categoryName' => new XML_RPC_Value($category_name[$new_keys[$i]], 'string'), 
                      'categoryId' => new XML_RPC_Value($new_keys[$i], 'string'),
                      'isPrimary' => new XML_RPC_Value($isPrimary, 'boolean')), "struct"));
    }

    $val = new XML_RPC_Value($ret, "array"); 
    return xrResp($val);
}

//setPostCategories  --- IDで指定された記事のカテゴリーを設定する。
//[引数]
//    string postid      : カテゴリー設定対象の記事ID。
//    string username    : blog のユーザ名(アカウント)。
//    string password    : パスワード。
//    array  categories  : カテゴリーの情報の配列。（以下の要素。）
//      string  categoryId : カテゴリID。
//      boolean isPrimary  : 当該カテゴリをプライマリーカテゴリに指定するか。
//[注意] isPrimary が指定されていなければ、先頭のカテゴリーがプライマリーとなる。
//[戻り値]
//    成功: true 、失敗: fault 。
function setPostCategories($params)
{
    global $xmlrpcerruser; // import user errcode value
    global $category_key, $category_name;

WriteLog("---\nIn setPostCategories\n");
    for ($i = 0; $i < $params->getNumParams(); $i++) {
        $val = $params->getParam($i);
        switch ($i) {
        case 0: //postid
            if (!getParamByType($postid, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "postid");
            break;
        case 1: //blog のユーザーアカウント
            if (!getParamByType($username, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "username");
            break;
        case 2: //上記アカウントのログインパスワード
            if (!getParamByType($password, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "password");
            break;
        case 3: //カテゴリーに関する構造体の配列
            if (!isset($val) || ($val->scalartyp() != "array")) {
                return xrErrResp(0, XR_ERR_ARG, "categories");
            }
            $categories = $val->getval();
            //if (count($categories) <= 0) return xrErrResp(0, XR_ERR_ARG, "categories");
            break;
        }
    }

    //許可されたユーザーかチェックする。
    $user = userCheck($username,$password);
    if (!$user) return xrErrResp(0, XR_ERR_USER, "");

    //プライマリーなカテゴリーを決定してそのキーを求める。
    $pri_index = -1;
    $cate_keys = array();//複数カテゴリーの設定もあり得るので、配列に取得しておく。
    for ($i = 0; $i < count($categories); $i++) {
        $elem = $categories[$i];
        $val = $elem['categoryId'];
        if (!isset($val) || ('string' != $val->scalartyp())) {
            return xrErrResp(0, XR_ERR_ARG, "categoryId");
        }
        $categoryId = $val->getval();
        $val = $elem['isPrimary'];
        if (!isset($val) || ('boolean' != $val->scalartyp())) {
            //moblog.uva.ne.jp が値を int で返してくるため。
            if (!isset($val) || ('int' != $val->scalartyp())) {
                return xrErrResp(0, XR_ERR_ARG, "isPrimary");
            }
        }
        $isPrimary  = $val->getval();
        array_push($cate_keys, $categoryId);
        if ($isPrimary) $pri_index = $i;
    }

    if ($pri_index >= 0) {
        $cate = $cate_keys[$pri_index];
    } else {
        //カテゴリーが設定されていない場合、そのユーザーの primarycategory にする。
        $cate = $user['primarycategory'];
        // primarycategory も設定されていない場合 fault 扱いにする。
        if (!$cate) return xrErrResp(0, XR_ERR_FUNCTION, "no primarycategory");
    }
    if (!setCategoryFromPostid($postid, $cate)) return xrErrResp(0, XR_ERR_FUNC, "setCategoryFromPostid");

WriteLog("Out setPostCategories\n");
    $val = new XML_RPC_Value(true, 'boolean'); 
    return xrResp($val);
}

//blogger_getUserInfo --- 指定ユーザーの情報を返す。
//[引数]
//    string appkey   : 任意の文字列でよい。
//    string username : blog のユーザ名(アカウント)。
//    string password : パスワード。
//[戻り値]
//    成功: 以下の要素の構造体を返す。
//        （ RFC では「投稿可能な weblog の数の分、構造体の配列を返す。」と
//           あるが、 MT で調べた結果の方に合わせて構造体を直に返す。）
//    string userid    : 当該 weblog でのユーザID。
//    string firstname : 名。
//    string lastname  : 姓。
//    string nickname  : ニックネーム。
//    string email     : e-mail address 。
//    string url       : 当該 weblog のトップページ URL。
//    失敗: fault 。
function blogger_getUserInfo($params)
{
    global $xmlrpcerruser; // import user errcode value
    global $category_key, $category_name;

    for ($i = 0; $i < $params->getNumParams(); $i++) {
        $val = $params->getParam($i);
        switch ($i) {
        case 0: //appkey(何でも良い。)
            if (!getParamByType($appkey, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "appkey");
            break;
        case 1: //blog のユーザーアカウント
            if (!getParamByType($username, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "username");
            break;
        case 2: //上記アカウントのログインパスワード
            if (!getParamByType($password, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "password");
            break;
        }
    }
    //許可されたユーザーかチェックする。
    $usr = userCheck($username,$password);
    if (!$usr) return xrErrResp(0, XR_ERR_USER, "");
    $ret = array('userid'  => new XML_RPC_Value($username, 'string'), 
                      'firstname'  => new XML_RPC_Value($usr['firstname'], 'string'), 
                      'lastname'  => new XML_RPC_Value($usr['lastname'], 'string'), 
                      'nickname'  => new XML_RPC_Value($usr['nickname'], 'string'), 
                      'email'  => new XML_RPC_Value($usr['email'], 'string'), 
                      'url'  => new XML_RPC_Value(SITE_URL, 'string'));
    $val = new XML_RPC_Value($ret, "struct"); 
    return xrResp($val);
}

//blogger_getUsersBlogs --- 投稿できる blog の配列を返す。
//[引数]
//    string appkey   : Movable Type, TypePad, ココログでは任意の文字列でよい。
//    string username : blog のユーザ名(アカウント)。
//    string password : パスワード。
//[戻り値]
//    成功: 投稿可能な weblog の数の分、以下の要素の構造体の配列を返す。
//    string url      : 当該 blog のトップページ URL。
//    string blogid   : 当該 blog の blog ID。
//    string blogName : 当該 blog 名称。
//    失敗: fault 。
function blogger_getUsersBlogs($params)
{
    global $xmlrpcerruser; // import user errcode value
    global $category_key, $category_name;

    for ($i = 0; $i < $params->getNumParams(); $i++) {
        $val = $params->getParam($i);
        switch ($i) {
        case 0: //appkey(何でも良い。)
            if (!getParamByType($appkey, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "appkey");
            break;
        case 1: //blog のユーザーアカウント
            if (!getParamByType($username, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "username");
            break;
        case 2: //上記アカウントのログインパスワード
            if (!getParamByType($password, $val, 'string')) return xrErrResp(0, XR_ERR_ARG, "password");
            break;
        }
    }
    //許可されたユーザーかチェックする。
    if (!userCheck($username,$password)) return xrErrResp(0, XR_ERR_USER, "");
    $ret = array();
    $blog_count = 1;//現在のところ、管理している blog は１つだけなので…。
    for ($i = 0; $i < $blog_count; $i++) {
        array_push($ret,
            new XML_RPC_Value(
                array('url'  => new XML_RPC_Value(SITE_URL, 'string'), 
                      'blogid'  => new XML_RPC_Value(SITE_BLOGID, 'string'), 
                      'blogName'  => new XML_RPC_Value(SITE_TITLE, 'string')), "struct"));
    }
    $val = new XML_RPC_Value($ret, "array"); 
    return xrResp($val);
}


//デバッグ用。
//$log = "\n---".date("Y-m-d H:i:s")."---\n".$HTTP_RAW_POST_DATA."\n---\n";
//WriteLog($log);

//サーバーのインスタンスを生成する。
$server  = new XML_RPC_Server(array(
                              'blogger.deletePost' => array('function' => 'blogger_deletePost'),
                              'blogger.getUserInfo' => array('function' => 'blogger_getUserInfo'),
                              'blogger.getUsersBlogs' => array('function' => 'blogger_getUsersBlogs'),
                              'mt.supportedMethods' => array('function' => 'mt_supportedMethods'),
                              'mt.publishPost' => array('function' => 'mt_publishPost'),
                              'mt.getCategoryList' => array('function' => 'mt_getCategoryList'),
                              'mt.getPostCategories' => array('function' => 'getPostCategories'),
                              'mt.setPostCategories' => array('function' => 'setPostCategories'),
                              'metaWeblog.getCategories' => array('function' => 'metaWeblog_getCategories'),
                              'metaWeblog.getPost' => array('function' => 'getPost'),
                              'metaWeblog.getRecentPosts' => array('function' => 'getRecentPosts'),
                              'metaWeblog.newMediaObject' => array('function' => 'newMediaObject'),
                              'metaWeblog.editPost' => array('function' => 'metaWeblog_editPost'),
                              'metaWeblog.newPost' => array('function' => 'newPost')
                              ),1,0);
?>

